# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
The render module defines a series of classes for interpreting and rendering
models in the `Wavefront object format`_.

.. note::

    All items in this module are available from the :mod:`picraft` namespace
    without having to import :mod:`picraft.render` directly.

.. _Wavefront object format: http://paulbourke.net/dataformats/obj/

The following items are defined in the module:


Model
=====

.. autoclass:: Model
    :members:


ModelFace
=========

.. autoclass:: ModelFace
    :members:

"""

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')


import io
import warnings
from collections import namedtuple, defaultdict
from itertools import chain

from .vector import Vector, vector_range, filled, lines
from .block import Block
from .exc import (
    UnsupportedCommand,
    NegativeWeight,
    )


COMMANDS = {
    'v',           # geometric vertices
    'vt',          # texture vertices
    'vn',          # normal vertices
    'vp',          # parameter space vertices

    'cstype',      # rational or non-rational curve or surface type
    'deg',         # degree
    'bmat',        # basis matrix
    'step',        # step size

    'p',           # point
    'l',           # line
    'f',           # face
    'curv',        # curve
    'curv2',       # 2D curve
    'surf',        # surface

    'parm',        # parameter values
    'trim',        # outer trimming loop
    'hole',        # inner trimming loop
    'scrv',        # special curve
    'sp',          # special point
    'end',         # end statement

    'con',         # connect

    'g',           # group
    's',           # smoothing group
    'mg',          # merging group
    'o',           # object name

    'bevel',       # bevel interpolation
    'c_interp',    # color interpolation
    'd_interp',    # dissolve interpolation
    'lod',         # level of detail
    'usemtl',      # material name
    'mtllib',      # material library
    'shadow_obj',  # shadow casting
    'trace_obj',   # ray tracing
    'ctech',       # curve approximation technique
    'stech',       # surface approximation technique

    'call',        # file inclusion
    'csh',         # shell execution

    # Obsolete
    'res',         # set number of segments in patches
    'bzp',         # bezier patch
    'bsp',         # b-spline patch
    'cdc',         # cardinal curve
    'cdp',         # cardinal patch
    }

IGNORED = COMMANDS - {'v', 'vn', 'vt', 'vp', 'f', 'g', 'usemtl'}


class Vertex(namedtuple('Vertex', ('x', 'y', 'z', 'w'))):
    """
    Represents a geometric vertex in a Wavefront obj file. The w component is
    optional and defaults to 1.0.
    """

    __slots__ = ()

    def __new__(cls, x, y, z, w=1.0):
        x = float(x)
        y = float(y)
        z = float(z)
        w = float(w)
        if w <= 0.0:
            warnings.warn(NegativeWeight('negative or zero weight: %f'))
        return super(Vertex, cls).__new__(cls, x, y, z, w)

    @property
    def __dict__(self):
        return super(Vertex, self).__dict__


class VertexParameter(namedtuple('VertexParameter', ('u', 'v', 'w'))):
    """
    Represents a point in the parameter space of a curve or surface. The w
    component is optional and defaults to 1.0.
    """

    __slots__ = ()

    def __new__(cls, u, v=0.0, w=1.0):
        u = float(u)
        v = float(v)
        w = float(w)
        return super(VertexParameter, cls).__new__(cls, u, v, w)

    @property
    def __dict__(self):
        return super(VertexParameter, self).__dict__


class VertexNormal(namedtuple('VertexNormal', ('i', 'j', 'k'))):
    """
    Represents a normal vector with components i, j, and k.
    """

    __slots__ = ()

    def __new__(cls, i, j, k):
        i = float(i)
        j = float(j)
        k = float(k)
        return super(VertexNormal, cls).__new__(cls, i, j, k)

    @property
    def __dict__(self):
        return super(VertexNormal, self).__dict__


class VertexTexture(namedtuple('VertexTexture', ('u', 'v', 'w'))):
    """
    Represents a texture vertex and its coordinates. The v and w components
    are optional and default to 0.0.
    """

    __slots__ = ()

    def __new__(cls, u, v=0.0, w=0.0):
        u = float(u)
        v = float(v)
        w = float(w)
        return super(VertexTexture, cls).__new__(cls, u, v, w)

    @property
    def __dict__(self):
        return super(VertexTexture, self).__dict__


class FaceIndex(namedtuple('FaceIndex', ('v', 'vt', 'vn'))):
    """
    Represents a vertex, optional texture, and optional normal reference in
    a face statement.
    """

    __slots__ = ()

    def __new__(cls, v, vt=None, vn=None):
        v = int(v)
        vt = vt if vt is None else int(vt)
        vn = vn if vn is None else int(vn)
        return super(FaceIndex, cls).__new__(cls, v, vt, vn)

    @classmethod
    def from_string(cls, s):
        s = s.split('/')
        v = s[0]
        vt = s[1] if len(s) > 1 else None
        vn = s[2] if len(s) > 2 else None
        if len(s) > 3:
            raise ValueError('too many values in face index')
        return cls.__new__(cls, v, vt, vn)

    @property
    def __dict__(self):
        return super(FaceIndex, self).__dict__


class FaceIndexes(object):
    """
    Represents a face containing an arbitrary number of vertices (>3). A
    :meth:`resolve` method is provided to permit resolution of negative indices
    but otherwise the resulting object is effectively an immutable list.
    """

    __slots__ = ('_items', '_material', '_groups')

    def __init__(self, *indexes):
        if len(indexes) < 3:
            raise ValueError('insufficient number of vertixes for face')
        self._items = [FaceIndex.from_string(i) for i in indexes]

    def __repr__(self):
        return '<FaceIndexes %d vertexes>' % len(self)

    def __len__(self):
        return len(self._items)

    def __iter__(self):
        return iter(self._items)

    def __getitem__(self, index):
        return self._items[index]


class Group(object):
    """
    Represents a group, or groups as a list of :attr:`names`. Any number of
    groups may be "active" at a time, and if no group names are specified,
    "default" is used.
    """

    __slots__ = ('_names')

    def __init__(self, *names):
        if not names:
            names = ['default']
        self._names = frozenset(names)

    def __repr__(self):
        return '<Group %s>' % ', '.join(repr(n) for n in self._names)

    @property
    def names(self):
        return self._names


class Material(str):
    """
    Represents a material (or more specifically a command to switch material).
    This is simply a derivative of Python's built-in :class:`str`.
    """

    def __new__(cls, *args):
        if len(args) == 0:
            raise ValueError('missing material name')
        return super(Material, cls).__new__(cls, args[0])

    def __repr__(self):
        return '<Material %s>' % super(Material, self).__repr__()


class Parser(object):
    """
    Parser for the Alias|Wavefront Obj format. Based partially on the
    `specification`_ published in Appendix B1 of the Alias|Wavefront manual.
    Handles backslash continuation of any lines, ignoring comments and blank
    lines, and assumes commands always start in the first column.

    The Parser acts as an iterable object (akin to a generator). Constructed
    with a filename or file-like object (which must return unicode strings in
    Python 3, as opposed to bytes), the instance acts as an iterable yielding
    :class:`Vertex`, :class:`VertexParameter`, :class:`VertexNormal`,
    :class:`VertexTexture`, :class:`FaceIndex`, :class:`Group`, and
    :class:`Material` instances depending on the statement encountered.

    ASCII encoding of the source file is assumed (no other character sets are
    supported), and all other legitimate commands are recognized but ignored.
    Such commands will cause an :exc:`UnsupportedCommand` warning to be raised
    but this is ignored by default.
    """
    def __init__(self, source):
        if isinstance(source, bytes):
            source = source.decode('utf-8')
        self._opened = isinstance(source, str)
        if self._opened:
            self._source = io.open(source, 'r', encoding='ascii')
        else:
            self._source = source

    def close(self):
        if self._opened:
            self._source.close()
        self._source = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()

    def __iter__(self):
        compound = []
        for line_num, line in enumerate(self._source, start=1):
            line = line.rstrip()
            if line.endswith('\\'):
                compound.append(line[:-1])
            else:
                compound.append(line)
                line = ' '.join(compound)
                compound = []
                if line and not line.startswith('#'):
                    try:
                        command, params = line.split(None, 1)
                    except ValueError:
                        command, params = line, ''
                    params = params.split()
                    if command in IGNORED:
                        warnings.warn(UnsupportedCommand(
                            'line %d: unsupported command %s' % (line_num, command)))
                    elif not command in COMMANDS:
                        raise ValueError(
                            'line %d: unknown command %s' % (line_num, command))
                    else:
                        yield {
                            'v':      Vertex,
                            'vn':     VertexNormal,
                            'vp':     VertexParameter,
                            'vt':     VertexTexture,
                            'f':      FaceIndexes,
                            'g':      Group,
                            'usemtl': Material,
                            }[command](*params)


class ModelFace(object):
    """
    Represents a face belonging to a :class:`Model`. A face consists of three
    or more :attr:`vectors` which are all `coplanar`_ (belonging to the same
    two-dimensional plane within the three-dimensional space).

    A face also has a :attr:`material`. As Minecraft's rendering is relatively
    crude this is simply stored as the name of the material; it is up to the
    user to map this to a meaningful block type. Finally each face belongs to
    zero or more :attr:`groups` which can be used to distinguish components of
    a model from each other.

    .. _coplanar: https://en.wikipedia.org/wiki/Coplanarity
    """

    def __init__(self, vectors, material, groups):
        self._vectors = tuple(vectors)
        self._groups = frozenset(groups)
        self._material = material

    @property
    def material(self):
        """
        The material assigned to the face. This is simply stored as the name of
        the material as it would be ridiculous to even attempt to emulate the
        material model of a full ray-tracer as Minecraft blocks.

        The :meth:`Model.render` method provides a simple means for mapping a
        material name to a block type in Minecraft.
        """
        return self._material

    @property
    def groups(self):
        """
        The set of groups that the face belongs to. By default all faces belong
        to a :class:`Model`. However, in additionl to this a face can belong to
        zero or more "groups" which are effectively components of the model.
        This facility can be used to render particular parts of a model.
        """
        return self._groups

    @property
    def vectors(self):
        """
        The sequence of vectors that makes up the face. These are assumed to be
        `coplanar`_ but this is not explicitly checked. Each point is
        represented as a :class:`~picraft.vector.Vector` instance.

        .. _coplanar: https://en.wikipedia.org/wiki/Coplanarity
        """
        return self._vectors

    def __repr__(self):
        return '<ModelFace %d points, material="%s", groups=%s>' % (
                len(self._vectors), self._material,
                '{%s}' % ', '.join('"%s"' % g for g in self._groups))


class Model(object):
    """
    Represents a three-dimensional model parsed from an Alias|Wavefront `object
    file`_ (.obj extension). The constructor accepts a *source* parameter which
    can be a filename or file-like object (in the latter case, this must be
    opened in text mode such that it returns unicode strings rather than bytes
    in Python 3).

    The optional *swap_yz* parameter specifies whether the Y and Z coordinates
    of each vertex in the model will be swapped; some models require this to
    render correctly in Minecraft, some do not.

    The :attr:`faces` attribute provides access to all object faces extracted
    from the file's content. The :attr:`materials` property enumerates all
    material names used by the object. The :attr:`groups` mapping maps group
    names to subsets of the available faces. The :attr:`bounds` attribute
    provides a range describing the bounding box of the unscaled model.

    Finally, the :meth:`render` method can be used to easily render the object
    in the Minecraft world at the specified scale, and with a given material
    mapping.

    .. _object file: https://en.wikipedia.org/wiki/Wavefront_.obj_file
    """

    def __init__(self, source, swap_yz=False):
        self._faces = []
        self._materials = set()
        self._groups = defaultdict(list)
        self._swap_yz = swap_yz
        self._parse(source)

    def _parse(self, source):
        vertexes = []
        textures = []
        normals = []
        active_groups = set()
        active_material = None
        for i in Parser(source):
            if isinstance(i, Vertex):
                vertexes.append(i)
            elif isinstance(i, VertexTexture):
                textures.append(i)
            elif isinstance(i, VertexNormal):
                normals.append(i)
            elif isinstance(i, Group):
                active_groups = i.names
            elif isinstance(i, Material):
                self._materials.add(i)
                active_material = i
            elif isinstance(i, FaceIndexes):
                if active_material is None:
                    self._materials.add(None)
                vectors = [
                    Vector(v.x, v.z, v.y) if self._swap_yz else
                    Vector(v.x, v.y, v.z)
                    for vi in i
                    for v in (vertexes[vi.v - 1 if vi.v > 0 else len(vertexes) + vi.v],)
                    ]
                face = ModelFace(vectors, active_material, active_groups)
                self._faces.append(face)
                for group in active_groups:
                    self._groups[group].append(face)

    @property
    def faces(self):
        """
        Returns the sequence of faces that make up the model. Each instance of
        this sequence is a :class:`ModelFace` instance which provides details
        of the coordinates of the face vertices, the face material, etc.
        """
        return self._faces

    @property
    def materials(self):
        """
        Returns the set of materials used by the model. This is derived from
        the :class:`~ModelFace.material` assigned to each face of the model.
        """
        return self._materials

    @property
    def groups(self):
        """
        A mapping of group names to sequences of :class:`ModelFace` instances.
        This can be used to extract a component of the model for further
        processing or rendering.
        """
        return self._groups

    @property
    def bounds(self):
        """
        Returns a vector range which completely encompasses the model at scale
        1.0. This can be useful for determining scaling factors when rendering.

        .. note::

            The bounding box returned is `axis-aligned`_ and is not guaranteed
            to be the minimal bounding box for the model.

        .. _axis-aligned: https://en.wikipedia.org/wiki/Minimum_bounding_box#Axis-aligned_minimum_bounding_box
        """
        min_v = Vector(
            min(v.x for f in self.faces for v in f.vectors),
            min(v.y for f in self.faces for v in f.vectors),
            min(v.z for f in self.faces for v in f.vectors),
            ).floor()
        max_v = Vector(
            max(v.x for f in self.faces for v in f.vectors),
            max(v.y for f in self.faces for v in f.vectors),
            max(v.z for f in self.faces for v in f.vectors),
            ).floor()
        return vector_range(min_v, max_v + 1)

    def render(self, scale=1.0, materials=None, groups=None):
        """
        Renders the model as a :class:`dict` mapping vectors to block types.
        Effectively this rounds the vertices of each face to integers (after
        applying the *scale* multiplier, which defaults to 1.0), then calls
        :func:`~picraft.vector.filled` and :func:`~picraft.vector.lines` to
        obtain the complete coordinates of each face.

        Each coordinate then needs to be mapped to a block type. By default
        the material name is simply passed to the :class:`~picraft.block.Block`
        constructor. This assumes that material names are valid Minecraft
        block types (see :attr:`~picraft.block.Block.NAMES`).

        You can override this mechanism with the *materials* parameter. This
        can be set to a mapping (e.g. a :class:`dict`) which maps material
        names to :class:`~picraft.block.Block` instances. For example::

            from picraft import Model, Block

            m = Model('airboat.obj')
            d = m.render(materials={
                'bluteal':  Block('diamond_block'),
                'bronze':   Block('gold_block'),
                'dkdkgrey': Block((64, 64, 64)),
                'dkteal':   Block('#000080'),
                'red':      Block('#ff0000'),
                'silver':   Block.from_color('#ffffff'),
                'black':    Block(id=35, data=15),
                None:       Block('stone'),
                })

        .. note::

            Some object files may include faces with no associated material.
            In this case you will need to map ``None`` to a block type, as in
            the example above.

        Alternatively, *materials* can be a callable which will be called with
        the :class:`ModelFace` being rendered, which should return a block
        type. The following is equivalent to the default behaviour::

            from picraft import Model, Block

            m = Model('airboat.obj')
            d = m.render(materials=lambda f: Block(f.material))

        If you simply want to preview a shape without bothering with any
        material mapping you can use this method to map any face to a single
        material (in this case stone)::

            from picraft import Model, Block

            m = Model('airboat.obj')
            d = m.render(materials=lambda f: Block('stone'))

        If the *materials* mapping or callable returns ``None`` instead of a
        :class:`~picraft.block.Block` instance, the corresponding blocks will
        not be included in the result. This is a simple mechanism for excluding
        parts of a model. The other mechanism for achieving this is the
        *groups* parameter which specifies which sub-components of the model
        should be rendered. This can be specified as a string (indicating that
        only that sub-component should be rendered) or as a sequence of strings
        (indicating that all specified sub-components should be rendered).

        The result is a mapping of :class:`~picraft.vector.Vector` to
        :class:`~picraft.block.Block` instances. Rendering the result in the
        main world should be as trivial as the following code::

            from picraft import World, Model

            w = World()
            m = Model('airboat.obj').render(scale=2.0)
            with w.connection.batch_start():
                for v, b in m.items():
                    w.blocks[v] = b

        Of course, you may choose to further transform the result first. This
        can be accomplished by modifying the vectors as you set them::

            from picraft import World, Model, Y

            w = World()
            m = Model('airboat.obj').render(scale=2.0)
            with w.connection.batch_start():
                for v, b in m.items():
                    w.blocks[v + 10*Y] = b

        Alternatively you may choose to use a dict-comprehension::

            from picraft import Model, Vector

            m = Model('airboat.obj').render(scale=2.0)
            offset = Vector(y=10)
            m = {v + offset: b for v, b in m.items()}

        Note that the Alias|Wavefront `object file`_ format is a relatively
        simple text based format that can be constructed by hand without
        too much difficulty. Combined with the default mapping of material
        names to block types, this enables another means of constructing
        objects in the Minecraft world. For example, see :ref:`models`.

        .. _object file: https://en.wikipedia.org/wiki/Wavefront_.obj_file
        """
        if materials is None:
            materials = lambda f: Block(f.material)
        if isinstance(groups, bytes):
            groups = groups.decode('utf-8')
        if groups is None:
            faces = self.faces
        elif isinstance(groups, str):
            faces = self.groups[groups]
        else:
            faces = chain(*(self.groups[g] for g in groups))
        result = {}
        for face in faces:
            try:
                b = materials[face.material]
            except KeyError:
                raise KeyError('missing mapping for material "%s"' % face.material)
            except TypeError:
                b = materials(face)
            if b is not None:
                points = ((p * scale).round() for p in face.vectors)
                for v in filled(lines(points)):
                    result[v] = b
        return result




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
The vector module defines the :class:`Vector` class, which is the usual method
of represent coordinates or vectors when dealing with the Minecraft world. It
also provides functions like :func:`vector_range` for generating sequences of
vectors.

.. note::

    All items in this module are available from the :mod:`picraft` namespace
    without having to import :mod:`picraft.vector` directly.

The following items are defined in the module:


Vector
======

.. autoclass:: Vector(x=0, y=0, z=0)


Short-hand variants
===================

The :class:`Vector` class is used sufficiently often to justify the inclusion
of some shortcuts. The class itself is also available as ``V``, and vectors
representing the three axes are each available as ``X``, ``Y``, and ``Z``.
Finally, a vector representing the origin is available as ``O``::

    >>> from picraft import V, O, X, Y, Z
    >>> O
    Vector(x=0, y=0, z=0)
    >>> 2 * X
    Vector(x=2, y=0, z=0)
    >>> X + Y
    Vector(x=1, y=1, z=0)
    >>> (X + Y).angle_between(X)
    45.00000000000001
    >>> V(3, 4, 5).projection(X)
    3.0
    >>> X.rotate(90, about=Y)
    Vector(x=0.0, y=0.0, z=1.0)


vector_range
============

.. autoclass:: vector_range
    :members:


line
====

.. autofunction:: line


lines
=====

.. autofunction:: lines


circle
======

.. autofunction:: circle


sphere
======

.. autofunction:: sphere


filled
======

.. autofunction:: filled
"""

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')
from .compat import range


import math
from functools import total_ordering
from collections import namedtuple, Sequence
try:
    from itertools import zip_longest, islice, tee
except ImportError:
    # Py2 compat
    from itertools import izip_longest as zip_longest, islice, tee


class Vector(namedtuple('Vector', ('x', 'y', 'z'))):
    """
    Represents a 3-dimensional vector.

    This :func:`~collections.namedtuple` derivative represents a 3-dimensional
    vector with :attr:`x`, :attr:`y`, :attr:`z` components. Instances can be
    constructed in a number of ways: by explicitly specifying the x, y, and z
    components (optionally with keyword identifiers), or leaving the empty to
    default to 0::

        >>> Vector(1, 1, 1)
        Vector(x=1, y=1, z=1)
        >>> Vector(x=2, y=0, z=0)
        Vector(x=2, y=0, z=0)
        >>> Vector()
        Vector(x=0, y=0, z=0)
        >>> Vector(y=10)
        Vector(x=0, y=10, z=0)

    Shortcuts are available for vectors representing the X, Y, and Z axes::

        >>> X
        Vector(x=1, y=0, z=0)
        >>> Y
        Vector(x=0, y=1, z=0)

    Note that vectors don't much care whether their components are integers,
    floating point values, or ``None``::

        >>> Vector(1.0, 1, 1)
        Vector(x=1.0, y=1, z=1)
        >>> Vector(2, None, None)
        Vector(x=2, y=None, z=None)

    The class supports simple arithmetic operations with other vectors such as
    addition and subtraction, along with multiplication and division, raising
    to powers, bit-shifting, and so on. Such operations are performed
    element-wise [1]_::

        >>> v1 = Vector(1, 1, 1)
        >>> v2 = Vector(2, 2, 2)
        >>> v1 + v2
        Vector(x=3, y=3, z=3)
        >>> v1 * v2
        Vector(x=2, y=2, z=2)

    Simple arithmetic operations with scalars return a new vector with that
    operation performed on all elements of the original. For example::

        >>> v = Vector()
        >>> v
        Vector(x=0, y=0, z=0)
        >>> v + 1
        Vector(x=1, y=1, z=1)
        >>> 2 * (v + 2)
        Vector(x=4, y=4, z=4)
        >>> Vector(y=2) ** 2
        Vector(x=0, y=4, z=0)

    Within the Minecraft world, the X,Z plane represents the ground, while the
    Y vector represents height.

    .. note::

        Note that, as a derivative of :func:`~collections.namedtuple`,
        instances of this class are immutable. That is, you cannot directly
        manipulate the :attr:`x`, :attr:`y`, and :attr:`z` attributes; instead
        you must create a new vector (for example, by adding two vectors
        together). The advantage of this is that vector instances can be
        members of :class:`set` or keys in a :class:`dict`.

    .. [1] I realize math purists will hate this (and demand that abs() should
       be magnitude and * should invoke matrix multiplication), but the
       element wise operations are sufficiently useful to warrant the
       short-hand syntax.

    .. automethod:: replace

    .. automethod:: ceil

    .. automethod:: floor

    .. automethod:: dot

    .. automethod:: cross

    .. automethod:: distance_to

    .. automethod:: angle_between

    .. automethod:: project

    .. automethod:: rotate

    .. attribute:: x

        The position or length of the vector along the X-axis. In the Minecraft
        world this can be considered to run left-to-right.

    .. attribute:: y

        The position or length of the vector along the Y-axis. In the Minecraft
        world this can be considered to run vertically up and down.

    .. attribute:: z

        The position or length of the vector along the Z-axis. In the Minecraft
        world this can be considered as depth (in or out of the screen).

    .. autoattribute:: magnitude

    .. autoattribute:: unit
    """

    __slots__ = ()

    def __new__(cls, x=0, y=0, z=0):
        return super(Vector, cls).__new__(cls, x, y, z)

    @classmethod
    def from_string(cls, s, type=int):
        x, y, z = s.split(',')
        return cls(type(x), type(y), type(z))

    @property
    def __dict__(self):
        # This is required to work around a subtle issue encountered in Python
        # 3.3 and above. In these versions (probably deliberately), the
        # __dict__ property is not inherited by namedtuple descendents
        return super(Vector, self).__dict__

    def __str__(self):
        return '%s,%s,%s' % (self.x, self.y, self.z)

    def __add__(self, other):
        try:
            return Vector(self.x + other.x, self.y + other.y, self.z + other.z)
        except AttributeError:
            return Vector(self.x + other, self.y + other, self.z + other)

    __radd__ = __add__

    def __sub__(self, other):
        try:
            return Vector(self.x - other.x, self.y - other.y, self.z - other.z)
        except AttributeError:
            return Vector(self.x - other, self.y - other, self.z - other)

    def __mul__(self, other):
        try:
            return Vector(self.x * other.x, self.y * other.y, self.z * other.z)
        except AttributeError:
            return Vector(self.x * other, self.y * other, self.z * other)

    __rmul__ = __mul__

    def __truediv__(self, other):
        try:
            return Vector(self.x / other.x, self.y / other.y, self.z / other.z)
        except AttributeError:
            return Vector(self.x / other, self.y / other, self.z / other)

    def __floordiv__(self, other):
        try:
            return Vector(self.x // other.x, self.y // other.y, self.z // other.z)
        except AttributeError:
            return Vector(self.x // other, self.y // other, self.z // other)

    def __mod__(self, other):
        try:
            return Vector(self.x % other.x, self.y % other.y, self.z % other.z)
        except AttributeError:
            return Vector(self.x % other, self.y % other, self.z % other)

    def __pow__(self, other, modulo=None):
        if modulo is not None:
            try:
                # XXX What about other vector, modulo scalar, and other scalar, modulo vector?
                return Vector(
                        pow(self.x, other.x, modulo.x),
                        pow(self.y, other.y, modulo.y),
                        pow(self.z, other.z, modulo.z))
            except AttributeError:
                return Vector(
                        pow(self.x, other, modulo),
                        pow(self.y, other, modulo),
                        pow(self.z, other, modulo))
        try:
            return Vector(
                    pow(self.x, other.x),
                    pow(self.y, other.y),
                    pow(self.z, other.z))
        except AttributeError:
            return Vector(
                    pow(self.x, other),
                    pow(self.y, other),
                    pow(self.z, other))

    def __lshift__(self, other):
        try:
            return Vector(self.x << other.x, self.y << other.y, self.z << other.z)
        except AttributeError:
            return Vector(self.x << other, self.y << other, self.z << other)

    def __rshift__(self, other):
        try:
            return Vector(self.x >> other.x, self.y >> other.y, self.z >> other.z)
        except AttributeError:
            return Vector(self.x >> other, self.y >> other, self.z >> other)

    def __and__(self, other):
        try:
            return Vector(self.x & other.x, self.y & other.y, self.z & other.z)
        except AttributeError:
            return Vector(self.x & other, self.y & other, self.z & other)

    def __xor__(self, other):
        try:
            return Vector(self.x ^ other.x, self.y ^ other.y, self.z ^ other.z)
        except AttributeError:
            return Vector(self.x ^ other, self.y ^ other, self.z ^ other)

    def __or__(self, other):
        try:
            return Vector(self.x | other.x, self.y | other.y, self.z | other.z)
        except AttributeError:
            return Vector(self.x | other, self.y | other, self.z | other)

    def __neg__(self):
        return Vector(-self.x, -self.y, -self.z)

    def __pos__(self):
        return self

    def __abs__(self):
        return Vector(abs(self.x), abs(self.y), abs(self.z))

    def __bool__(self):
        return bool(self.x or self.y or self.z)

    def __trunc__(self):
        return Vector(math.trunc(self.x), math.trunc(self.y), math.trunc(self.z))

    # Py2 compat
    __nonzero__ = __bool__
    __div__ = __truediv__

    def replace(self, x=None, y=None, z=None):
        """
        Return the vector with the x, y, or z axes replaced with the specified
        values. For example::

            >>> Vector(1, 2, 3).replace(z=4)
            Vector(x=1, y=2, z=4)
        """
        # XXX What if I want to use None?
        return Vector(
            self.x if x is None else x,
            self.y if y is None else y,
            self.z if z is None else z)

    def floor(self):
        """
        Return the vector with the floor of each component. This is only useful
        for vectors containing floating point components::

            >>> Vector(0.5, -0.5, 1.9)
            Vector(0.0, -1.0, 1.0)
        """
        return Vector(
            int(math.floor(self.x)),
            int(math.floor(self.y)),
            int(math.floor(self.z)))

    def ceil(self):
        """
        Return the vector with the ceiling of each component. This is only
        useful for vectors containing floating point components::

            >>> Vector(0.5, -0.5, 1.2)
            Vector(1.0, 0.0, 2.0)
        """
        return Vector(
            int(math.ceil(self.x)),
            int(math.ceil(self.y)),
            int(math.ceil(self.z)))

    def round(self, ndigits=0):
        """
        Return the vector with the rounded value of each component. This is
        only useful for vectors containing floating point components::

            >>> Vector(0.5, -0.5, 1.2)
            Vector(1.0, -1.0, 1.0)

        The *ndigits* argument operates as it does in the built-in
        :func:`round` function, specifying the number of decimal (or integer)
        places to round to.
        """
        if ndigits <= 0:
            return Vector(
                int(round(self.x, ndigits)),
                int(round(self.y, ndigits)),
                int(round(self.z, ndigits)))
        else:
            return Vector(
                round(self.x, ndigits),
                round(self.y, ndigits),
                round(self.z, ndigits))

    def dot(self, other):
        """
        Return the `dot product`_ of the vector with the *other* vector. The
        result is a scalar value. For example::

            >>> Vector(1, 2, 3).dot(Vector(2, 2, 2))
            12
            >>> Vector(1, 2, 3).dot(X)
            1

        .. _dot product: http://en.wikipedia.org/wiki/Dot_product
        """
        return self.x * other.x + self.y * other.y + self.z * other.z

    def cross(self, other):
        """
        Return the `cross product`_ of the vector with the *other* vector. The
        result is another vector. For example::

            >>> Vector(1, 2, 3).cross(Vector(2, 2, 2))
            Vector(x=-2, y=4, z=-2)
            >>> Vector(1, 2, 3).cross(X)
            Vector(x=0, y=3, z=-2)

        .. _cross product: http://en.wikipedia.org/wiki/Cross_product
        """
        return Vector(
                self.y * other.z - self.z * other.y,
                self.z * other.x - self.x * other.z,
                self.x * other.y - self.y * other.x)

    def distance_to(self, other):
        """
        Return the Euclidian distance between two three dimensional points
        (represented as vectors), calculated according to `Pythagoras'
        theorem`_. For example::

            >>> Vector(1, 2, 3).distance_to(Vector(2, 2, 2))
            1.4142135623730951
            >>> Vector().distance_to(X)
            1.0

        .. _Pythagoras' theorem: http://en.wikipedia.org/wiki/Pythagorean_theorem
        """
        return (other - self).magnitude

    def angle_between(self, other):
        """
        Returns the angle between this vector and the *other* vector on a plane
        that contains both vectors. The result is measured in degrees. For
        example::

            >>> X.angle_between(Y)
            90.0
            >>> (X + Y).angle_between(X)
            45.00000000000001
        """
        return math.degrees(math.acos(self.unit.dot(other.unit)))

    def project(self, other):
        """
        Return the `scalar projection`_ of this vector onto the *other* vector.
        This is a scalar indicating the length of this vector in the direction
        of the *other* vector. For example::

            >>> Vector(1, 2, 3).project(2 * Y)
            2.0
            >>> Vector(3, 4, 5).project(Vector(3, 4, 0))
            5.0

        .. _scalar projection: https://en.wikipedia.org/wiki/Scalar_projection
        """
        return self.dot(other.unit)

    def rotate(self, angle, about, origin=None):
        """
        Return this vector after `rotation`_ of *angle* degrees about the line
        passing through *origin* in the direction *about*. Origin defaults to
        the vector 0, 0, 0. Hence, if this parameter is omitted this method
        calculates rotation about the axis (through the origin) defined by
        *about*.  For example::

            >>> Y.rotate(90, about=X)
            Vector(x=0, y=6.123233995736766e-17, z=1.0)
            >>> Vector(3, 4, 5).rotate(30, about=X, origin=10 * Y)
            Vector(x=3.0, y=2.3038475772933684, z=1.330127018922194)

        Information about rotation around arbitrary lines was obtained from
        `Glenn Murray's informative site`_.

        .. _rotation: https://en.wikipedia.org/wiki/Rotation_group_SO%283%29
        .. _Glenn Murray's informative site: http://inside.mines.edu/fs_home/gmurray/ArbitraryAxisRotation/
        """
        r = math.radians(angle)
        sin = math.sin(r)
        cos = math.cos(r)
        x, y, z = self
        if origin is None:
            # Fast-paths: rotation about a specific unit axis
            if about == X:
                return Vector(x, y * cos - z * sin, y * sin + z * cos)
            elif about == Y:
                return Vector(z * sin + x * cos, y, z * cos - x * sin)
            elif about == Z:
                return Vector(x * cos - y * sin, x * sin + y * cos, z)
            elif about == negX:
                return Vector(x, y * cos + z * sin, z * cos - y * sin)
            elif about == negY:
                return Vector(x * cos - z * sin, y, z * cos + x * sin)
            elif about == negZ:
                return Vector(x * cos + y * sin, y * cos - x * sin, z)
            # Rotation about an arbitrary axis
            u, v, w = about.unit
            return Vector(
                u * (u * x + v * y + w * z) * (1 - cos) + x * cos + (-w * y + v * z) * sin,
                v * (u * x + v * y + w * z) * (1 - cos) + y * cos + ( w * x - u * z) * sin,
                w * (u * x + v * y + w * z) * (1 - cos) + z * cos + (-v * x + u * y) * sin)
        # Rotation about an arbitrary line
        a, b, c = origin
        u, v, w = about.unit
        return Vector(
            (a * (v ** 2 + w ** 2) - u * (b * v + c * w - u * x - v * y - w * z)) * (1 - cos) + x * cos + (-c * v + b * w - w * y + v * z) * sin,
            (b * (u ** 2 + w ** 2) - v * (a * u + c * w - u * x - v * y - w * z)) * (1 - cos) + y * cos + ( c * u - a * w + w * x - u * z) * sin,
            (c * (u ** 2 + v ** 2) - w * (a * u + b * v - u * x - v * y - w * z)) * (1 - cos) + z * cos + (-b * u + a * v - v * x + u * y) * sin)

    @property
    def magnitude(self):
        """
        Returns the magnitude of the vector. This could also be considered the
        distance of the vector from the origin, i.e. ``v.magnitude`` is
        equivalent to ``Vector().distance_to(v)``. For example::

            >>> Vector(2, 4, 4).magnitude
            6.0
            >>> Vector().distance_to(Vector(2, 4, 4))
            6.0
        """
        return math.sqrt(self.x ** 2 + self.y ** 2 + self.z ** 2)

    @property
    def unit(self):
        """
        Return a `unit vector`_ (a vector with a magnitude of one) with the
        same direction as this vector::

            >>> X.unit
            Vector(x=1.0, y=0.0, z=0.0)
            >>> (2 * Y).unit
            Vector(x=0.0, y=1.0, z=0.0)

        .. note::

            If the vector's magnitude is zero, this property returns the
            original vector.

        .. _unit vector: http://en.wikipedia.org/wiki/Unit_vector
        """
        try:
            return self / self.magnitude
        except ZeroDivisionError:
            return self


# Short-hand variants
V = Vector
O = V()
X = V(x=1)
Y = V(y=1)
Z = V(z=1)
# These aren't exposed as short-hands; they're only pre-calculated here to
# speed up the fast-paths in the rotate() method
negX = V(x=-1)
negY = V(y=-1)
negZ = V(z=-1)


# TODO Yes, I'm being lazy with total_ordering ... probably ought to define all
# six comparison methods but I haven't got time right now ...

@total_ordering
class vector_range(Sequence):
    """
    Like :func:`range`, :class:`vector_range` is actually a type which
    efficiently represents a range of vectors. The arguments to the constructor
    must be :class:`Vector` instances (or objects which have integer ``x``,
    ``y``, and ``z`` attributes).

    If *step* is omitted, it defaults to ``Vector(1, 1, 1)``. If the *start*
    argument is omitted, it defaults to ``Vector(0, 0, 0)``. If any element
    of the *step* vector is zero, :exc:`ValueError` is raised.

    The contents of the range are largely determined by the *step* and *order*
    which specifies the order in which the axes of the range will be
    incremented.  For example, with the order ``'xyz'``, the X-axis will be
    incremented first, followed by the Y-axis, and finally the Z-axis. So, for
    a range with the default *start*, *step*, and *stop* set to ``Vector(3, 3,
    3)``, the contents of the range will be::

        >>> list(vector_range(Vector(3, 3, 3), order='xyz'))
        [Vector(0, 0, 0), Vector(1, 0, 0), Vector(2, 0, 0),
         Vector(0, 1, 0), Vector(1, 1, 0), Vector(2, 1, 0),
         Vector(0, 2, 0), Vector(1, 2, 0), Vector(2, 2, 0),
         Vector(0, 0, 1), Vector(1, 0, 1), Vector(2, 0, 1),
         Vector(0, 1, 1), Vector(1, 1, 1), Vector(2, 1, 1),
         Vector(0, 2, 1), Vector(1, 2, 1), Vector(2, 2, 1),
         Vector(0, 0, 2), Vector(1, 0, 2), Vector(2, 0, 2),
         Vector(0, 1, 2), Vector(1, 1, 2), Vector(2, 1, 2),
         Vector(0, 2, 2), Vector(1, 2, 2), Vector(2, 2, 2)]

    Vector ranges implement all common sequence operations except concatenation
    and repetition (due to the fact that range objects can only represent
    sequences that follow a strict pattern and repetition and concatenation
    usually cause the resulting sequence to violate that pattern).

    Vector ranges are extremely efficient compared to an equivalent
    :func:`list` or :func:`tuple` as they take a small (fixed) amount of
    memory, storing only the arguments passed in its construction and
    calculating individual items and sub-ranges as requested.

    Vector range objects implement the :class:`collections.Sequence` ABC,
    and provide features such as containment tests, element index lookup,
    slicing and support for negative indices.

    The default order (``'zxy'``) may seem an odd choice. This is primarily
    used as it's the order used by the Raspberry Juice server when returning
    results from the :ref:`world.getBlocks` call. In turn, Raspberry Juice
    probably uses this order as it results in returning a horizontal layer of
    vectors at a time (given the Y-axis is used for height in the Minecraft
    world).

    .. warning::

        Bear in mind that the ordering of a vector range may have affect tests
        for its ordering and equality. Two ranges with different orders are
        unlikely to test equal even though they may have the same *start*,
        *stop*, and *step* attributes (and thus contain the same vectors, but
        in a different order).

    Vector ranges can be accessed by integer index, by :class:`Vector` index,
    or by a slice of vectors. For example::

        >>> v = vector_range(Vector() + 1, Vector() + 3)
        >>> list(v)
        [Vector(x=1, y=1, z=1),
         Vector(x=1, y=1, z=2),
         Vector(x=2, y=1, z=1),
         Vector(x=2, y=1, z=2),
         Vector(x=1, y=2, z=1),
         Vector(x=1, y=2, z=2),
         Vector(x=2, y=2, z=1),
         Vector(x=2, y=2, z=2)]
        >>> v[0]
        Vector(x=1, y=1, z=1)
        >>> v[Vector(0, 0, 0)]
        Vector(x=1, y=1, z=1)
        >>> v[Vector(1, 0, 0)]
        Vector(x=2, y=1, z=1)
        >>> v[-1]
        Vector(x=2, y=2, z=2)
        >>> v[Vector() - 1]
        Vector(x=2, y=2, z=2)
        >>> v[Vector(x=1):]
        vector_range(Vector(x=2, y=1, z=1), Vector(x=3, y=3, z=3),
                Vector(x=1, y=1, z=1), order='zxy')
        >>> list(v[Vector(x=1):])
        [Vector(x=2, y=1, z=1),
         Vector(x=2, y=1, z=2),
         Vector(x=2, y=2, z=1),
         Vector(x=2, y=2, z=2)]

    However, integer slices are not currently permitted.
    """

    def __init__(
            self, start, stop=None, step=None, order='zxy'):
        if stop is None:
            start, stop = Vector(), start
        if step is None:
            step = Vector(1, 1, 1)
        if (start != start // 1) or (stop != stop // 1) or (step != step // 1):
            raise TypeError('integer vectors are required')
        if order not in ('xyz', 'xzy', 'yxz', 'yzx', 'zxy', 'zyx'):
            raise ValueError('invalid order: %s' % order)
        if not (step.x and step.y and step.z):
            raise ValueError('no element of step may be zero')
        self._start = start
        self._stop = stop
        self._step = step
        self._order = order
        self._ranges = [
            range(
                getattr(self.start, axis),
                getattr(self.stop, axis),
                getattr(self.step, axis))
            for axis in order
            ]
        self._indexes = [
            order.index(axis)
            for axis in 'xyz'
            ]
        self._xrange, self._yrange, self._zrange = (
            self._ranges[i] for i in self._indexes
            )
        self._len = len(self._xrange) * len(self._yrange) * len(self._zrange)

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop

    @property
    def step(self):
        return self._step

    @property
    def order(self):
        return self._order

    def __repr__(self):
        if self.start == Vector() and self.step == Vector() + 1:
            return 'vector_range(%r, order=%r)' % (self.stop, self.order)
        elif self.step == Vector() + 1:
            return 'vector_range(%r, %r, order=%r)' % (
                    self.start, self.stop, self.order)
        else:
            return 'vector_range(%r, %r, %r, order=%r)' % (
                    self.start, self.stop, self.step, self.order)

    def __len__(self):
        return self._len

    def __lt__(self, other):
        for v1, v2 in zip_longest(self, other):
            if v1 < v2:
                return True
            elif v1 > v2:
                return False
        return False

    def __eq__(self, other):
        # Fast-path: if the other object is an identical vector_range we
        # can quickly test whether we're equal
        if isinstance(other, vector_range):
            return (
                    self._xrange == other._xrange and
                    self._yrange == other._yrange and
                    self._zrange == other._zrange and
                    self.order == other.order
                    )
        # Normal case: test every element in each sequence
        for v1, v2 in zip_longest(self, other):
            if v1 != v2:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __reversed__(self):
        for i in reversed(range(len(self))):
            yield self[i]

    def __contains__(self, value):
        try:
            self.index(value)
        except ValueError:
            return False
        else:
            return True

    def __bool__(self):
        return len(self) > 0

    # Py2 compat
    __nonzero__ = __bool__

    def __getitem__(self, index):
        if isinstance(index, slice):
            return self._get_slice(index)
        elif isinstance(index, Vector):
            try:
                return Vector(*(
                    self._ranges[i][j]
                    for i, j in zip(self._indexes, index)
                    ))
            except IndexError:
                raise IndexError('list index out of range')
        else:
            if index < 0:
                index += len(self)
            if not (0 <= index < len(self)):
                raise IndexError('list index out of range')
            v = (
                self._ranges[0][index % len(self._ranges[0])],
                self._ranges[1][(index // len(self._ranges[0])) % len(self._ranges[1])],
                self._ranges[2][index // (len(self._ranges[0]) * len(self._ranges[1]))],
                )
            return Vector(*(v[i] for i in self._indexes))

    def _get_slice(self, s):
        try:
            step = Vector() + 1 if s.step is None else s.step
            start = Vector(None, None, None) if s.start is None else s.start
            stop = Vector(None, None, None) if s.stop is None else s.stop
            if not (step.x and step.y and step.z):
                raise ValueError(
                    "every element of the slice's step must be non-zero")
            x_range = self._xrange[slice(start.x, stop.x, step.x)]
            y_range = self._yrange[slice(start.y, stop.y, step.y)]
            z_range = self._zrange[slice(start.z, stop.z, step.z)]
        except AttributeError:
            raise ValueError(
                "vector_range slices must be composed of Vectors")
        return vector_range(
            Vector(x_range.start, y_range.start, z_range.start),
            Vector(x_range.stop, y_range.stop, z_range.stop),
            Vector(x_range.step, y_range.step, z_range.step),
            self.order)

    def index(self, value):
        """
        Return the zero-based index of *value* within the range, or raise
        :exc:`ValueError` if *value* does not exist in the range.
        """
        ranges = self._ranges
        i, j, k = (getattr(value, axis) for axis in self.order)
        try:
            i_indexes = set(rmod(len(ranges[0]), ranges[0].index(i), range(len(self))))
            j_indexes = set(
                    b
                    for a in rmod(len(ranges[1]), ranges[1].index(j),
                        range(len(self) // len(ranges[0])))
                    for b in rdiv(len(ranges[0]), a)
                    )
            k_indexes = set(rdiv(len(ranges[0]) * len(ranges[1]), ranges[2].index(k)))
            result = i_indexes & j_indexes & k_indexes
            assert len(result) == 1
            result = next(iter(result))
        except ValueError:
            raise ValueError('%r is not in range' % (value,))
        else:
            return result

    def count(self, value):
        """
        Return the count of instances of *value* within the range (note this
        can only be 0 or 1 in the case of a range, and thus is equivalent to
        testing membership with ``in``).
        """
        # count is provided by the ABC but inefficiently; given no vectors in
        # the range can be duplicated we provide a more efficient version here
        if value in self:
            return 1
        else:
            return 0


def rmod(denom, result, num_range):
    """
    Calculates the inverse of a mod operation.

    The *denom* parameter specifies the denominator of the original mod (%)
    operation. In this implementation, *denom* must be greater than 0. The
    *result* parameter specifies the result of the mod operation. For obvious
    reasons this value must be in the range ``[0, denom)`` (greater than or
    equal to zero and less than the denominator).

    Finally, *num_range* specifies the range that the numerator of the original
    mode operation can lie in. This must be an iterable sorted in ascending
    order with unique values (e.g. most typically a :func:`range`).

    The function returns the set of potential numerators (guaranteed to be a
    subset of *num_range*).
    """
    if denom <= 0:
        raise ValueError('invalid denominator')
    if not (0 <= result < denom):
        return set()
    if len(num_range) == 0:
        return set()
    assert num_range[-1] >= num_range[0]
    start = num_range[0] + (result - num_range[0] % denom) % denom
    stop = num_range.stop
    return range(start, stop, denom)


def rdiv(denom, result):
    """
    Calculates the inverse of a div operation.

    The *denom* parameter specifies the denominator of the original div (//)
    operation. In this implementation, *denom* must be greater than 0. The
    *result* parameter specifies the result of the div operation.

    The function returns the set of potential numerators.
    """
    if denom <= 0:
        raise ValueError('invalid denominator')
    return range(result * denom, result * denom + denom)


def sign(v):
    """
    Returns the sign of v as -1, 0, or 1; works for scalar values or
    :class:`Vector` instances.
    """
    try:
        return Vector(sign(v.x), sign(v.y), sign(v.z))
    except AttributeError:
        return 1 if v > 0 else -1 if v < 0 else 0


def line(start, end):
    """
    Generates the coordinates of a line joining the *start* and *end*
    :class:`Vector` instances inclusive. This is a generator function; points
    are yielded from *start*, proceeding to *end*. If you don't require all
    points you may terminate the generator at any point.

    For example::

        >>> list(line(O, V(10, 5, 0)))
        [Vector(x=0, y=0, z=0),
         Vector(x=1, y=1, z=0),
         Vector(x=2, y=1, z=0),
         Vector(x=3, y=2, z=0),
         Vector(x=4, y=2, z=0),
         Vector(x=5, y=3, z=0),
         Vector(x=6, y=3, z=0),
         Vector(x=7, y=4, z=0),
         Vector(x=8, y=4, z=0),
         Vector(x=9, y=5, z=0),
         Vector(x=10, y=5, z=0)]

    To draw the resulting line you can simply assign a block to the collection
    of vectors generated (or assign a sequence of blocks of equal length if you
    want the line to have varying block types)::

        >>> world.blocks[line(O, V(10, 5, 0))] = Block('stone')

    This is a three-dimensional implementation of `Bresenham's line
    algorithm`_, derived largely from `Bob Pendelton's implementation`_ (public
    domain).

    .. _Bresenham's line algorithm: https://en.wikipedia.org/wiki/Bresenham%27s_line_algorithm
    .. _Bob Pendelton's implementation: ftp://ftp.isc.org/pub/usenet/comp.sources.unix/volume26/line3d
    """
    delta = end - start
    # Calculate the amount to increment each axis by; only the dominant axis
    # will advance by this amount on *every* iteration. Other axes will only
    # increment when the error demands it
    pos_inc = sign(delta)
    # Set up a vector containing the error incrementor. This will be added to
    # values tracking the axis error on each iteration
    error_inc = abs(delta) << 1
    # Calculate the subordinate and dominant axes. The dominant axis is simply
    # the one in which we must move furthest
    sub_axis1, sub_axis2, dominant_axis = sorted(
            'xyz', key=lambda axis: getattr(error_inc, axis))
    # Set up the error decrementor. This will be subtracted from the error
    # values when they turn positive (indicating that the corresponding axis
    # should advance)
    error_dec = getattr(error_inc, dominant_axis)
    # Set up a vector to track the error (this is only really required for the
    # subordinate axes)
    error = error_inc - (error_dec >> 1)
    # Convert vectors to dicts for the remainder of the algorithm
    error = error._asdict()
    error_inc = error_inc._asdict()
    pos_inc = pos_inc._asdict()
    pos = start._asdict()
    end = getattr(end, dominant_axis)
    while True:
        yield Vector(**pos)
        if pos[dominant_axis] == end:
            break
        pos[dominant_axis] += pos_inc[dominant_axis]
        if error[sub_axis1] >= 0:
            pos[sub_axis1] += pos_inc[sub_axis1]
            error[sub_axis1] -= error_dec
        error[sub_axis1] += error_inc[sub_axis1]
        if error[sub_axis2] >= 0:
            pos[sub_axis2] += pos_inc[sub_axis2]
            error[sub_axis2] -= error_dec
        error[sub_axis2] += error_inc[sub_axis2]


def lines(points, closed=True):
    """
    Generator function which extends the :func:`line` function; this yields all
    vectors necessary to render the lines connecting the specified *points*
    (which is an iterable of :class:`Vector` instances).

    If the optional *closed* parameter is ``True`` (the default) the last point
    in the *points* sequence will be connected to the first point. Otherwise,
    the lines will be left disconnected (assuming the last point is not
    coincident with the first). For example::

        >>> points = [O, 4*X, 4*Z]
        >>> list(lines(points))
        [Vector(x=0, y=0, z=0),
         Vector(x=1, y=0, z=0),
         Vector(x=2, y=0, z=0),
         Vector(x=3, y=0, z=0),
         Vector(x=4, y=0, z=0),
         Vector(x=3, y=0, z=1),
         Vector(x=2, y=0, z=2),
         Vector(x=1, y=0, z=3),
         Vector(x=0, y=0, z=4),
         Vector(x=0, y=0, z=3),
         Vector(x=0, y=0, z=2),
         Vector(x=0, y=0, z=1),
         Vector(x=0, y=0, z=0)]

    To draw the resulting polygon you can simply assign a block to the
    collection of vectors generated (or assign a sequence of blocks of equal
    length if you want the polygon to have varying block types)::

        >>> world.blocks[lines(points)] = Block('stone')

    To generate the coordinates of a filled polygon, see the :func:`filled`
    function.
    """
    first = None
    start = None
    point = None
    for point in points:
        if start is None:
            first = point
            yield first
        else:
            for v in islice(line(start, point), 1, None):
                yield v
        start = point
    if point is None:
        raise ValueError('no points specified')
    if closed and first != point:
        for v in islice(line(point, first), 1, None):
            yield v


def circle(center, radius, plane=Y):
    """
    Generator function which yields the coordinates of a three-dimensional
    circle centered at the :class:`Vector` *center*. The *radius* parameter is
    a vector specifying the distance of the circumference from the center. The
    optional *plane* parameter (which defaults to the Y unit vector) specifies
    another vector which, in combination with the *radius* vector, gives the
    plane that the circle exists within.

    For example, to generate the coordinates of a circle centered at (0, 10,
    0), with a radius of 5 units, existing in the X-Y plane::

        >>> list(circle(O, 5*X))
        [Vector(x=-5, y=0, z=0), Vector(x=-5, y=1, z=0), Vector(x=-4, y=2, z=0),
         Vector(x=-4, y=3, z=0), Vector(x=-5, y=-1, z=0), Vector(x=-4, y=-2, z=0),
         Vector(x=-4, y=-3, z=0), Vector(x=-3, y=4, z=0), Vector(x=-3, y=-4, z=0),
         Vector(x=-2, y=4, z=0), Vector(x=-2, y=-4, z=0), Vector(x=-1, y=4, z=0),
         Vector(x=-1, y=-4, z=0), Vector(x=0, y=5, z=0), Vector(x=0, y=-5, z=0),
         Vector(x=1, y=4, z=0), Vector(x=1, y=-4, z=0), Vector(x=2, y=4, z=0),
         Vector(x=2, y=-4, z=0), Vector(x=3, y=4, z=0), Vector(x=3, y=-4, z=0),
         Vector(x=4, y=3, z=0), Vector(x=4, y=-3, z=0), Vector(x=4, y=2, z=0),
         Vector(x=5, y=1, z=0), Vector(x=5, y=0, z=0), Vector(x=4, y=-2, z=0),
         Vector(x=5, y=-1, z=0)]

    To generate another set of coordinates with the same center and radius, but
    existing in the X-Z (ground) plane::

        >>> list(circle(O, 5*X, plane=Z))
        [Vector(x=-5, y=0, z=0), Vector(x=-5, y=0, z=1), Vector(x=-4, y=0, z=2),
         Vector(x=-4, y=0, z=3), Vector(x=-5, y=0, z=-1), Vector(x=-4, y=0, z=-2),
         Vector(x=-4, y=0, z=-3), Vector(x=-3, y=0, z=4), Vector(x=-3, y=0, z=-4),
         Vector(x=-2, y=0, z=4), Vector(x=-2, y=0, z=-4), Vector(x=-1, y=0, z=4),
         Vector(x=-1, y=0, z=-4), Vector(x=0, y=0, z=5), Vector(x=0, y=0, z=-5),
         Vector(x=1, y=0, z=4), Vector(x=1, y=0, z=-4), Vector(x=2, y=0, z=4),
         Vector(x=2, y=0, z=-4), Vector(x=3, y=0, z=4), Vector(x=3, y=0, z=-4),
         Vector(x=4, y=0, z=3), Vector(x=4, y=0, z=-3), Vector(x=4, y=0, z=2),
         Vector(x=5, y=0, z=1), Vector(x=5, y=0, z=0), Vector(x=4, y=0, z=-2),
         Vector(x=5, y=0, z=-1)]

    To draw the resulting circle you can simply assign a block to the
    collection of vectors generated (or assign a sequence of blocks of equal
    length if you want the circle to have varying block types)::

        >>> world.blocks[circle(O, 5*X)] = Block('stone')

    The algorithm used by this function is based on a straight-forward
    differences of roots method, extended to three dimensions. This produces
    `worse looking`_ circles than the `midpoint circle algorithm`_ (also known
    as a the Bresenham circle algorithm), but isn't restricted to working in a
    simple cartesian plane.

    .. note::

        If you know of a three dimensional generalization of the midpoint
        circle algorithm (which handles entirely arbitrary planes, not merely
        simple X-Y, X-Z, etc. planes), please contact the `author`_!

    To create a filled circle, see the :func:`filled` function.

    .. _worse looking: https://sites.google.com/site/ruslancray/lab/projects/bresenhamscircleellipsedrawingalgorithm/bresenham-s-circle-ellipse-drawing-algorithm
    .. _midpoint circle algorithm: https://en.wikipedia.org/wiki/Midpoint_circle_algorithm
    .. _author: mailto:dave@waveform.org.uk
    """
    try:
        if radius.angle_between(plane) != 90:
            plane = radius.cross(-(radius.cross(plane)))
    except AttributeError:
        raise ValueError('radius must be a Vector instance')
    perp = plane.unit
    r = radius.magnitude**2
    last_points = None
    result = set()
    result_len = 0
    for radial_point in line(-radius, radius):
        circum_v = (perp * math.sqrt(r - radial_point.magnitude**2)).floor()
        top_point = (radial_point + circum_v)
        bottom_point = (radial_point - circum_v)
        if last_points is not None:
            top_last, bottom_last = last_points
            for p in line(top_last, top_point):
                p += center
                result.add(p)
                if len(result) > result_len:
                    result_len += 1
                    yield p
            for p in line(bottom_last, bottom_point):
                p += center
                result.add(p)
                if len(result) > result_len:
                    result_len += 1
                    yield p
        last_points = top_point, bottom_point


def sphere(center, radius):
    """
    Generator function which yields the coordinates of a hollow sphere. The
    *center* :class:`Vector` specifies the center of the sphere, and *radius*
    is a scalar number of blocks giving the distance from the center to the
    edge of the sphere.

    For example to create the coordinates of a sphere centered at the origin
    with a radius of 5 units::

        >>> list(sphere(O, 5))

    To draw the resulting sphere you can simply assign a block to the
    collection of vectors generated (or assign a sequence of blocks of equal
    length if you want the sphere to have varying block types)::

        >>> world.blocks[sphere(O, 5)] = Block('stone')

    The algorithm generates concentric circles covering the sphere's surface,
    advancing along the X, Y, and Z axes with duplicate elimination to prevent
    repeated coordinates being yielded. Three axes are required to eliminate
    gaps in the surface.
    """
    r = radius**2
    last_points = None
    result = set()
    result_len = 0
    for radial_point in range(-radius, radius + 1):
        circum_dist = int(round(math.sqrt(r - radial_point**2)))
        for p in circle(X * radial_point, Z * circum_dist):
            p += center
            result.add(p)
            if len(result) > result_len:
                result_len += 1
                yield p
        for p in circle(Y * radial_point, Z * circum_dist, plane=X):
            p += center
            result.add(p)
            if len(result) > result_len:
                result_len += 1
                yield p
        for p in circle(Z * radial_point, X * circum_dist):
            p += center
            result.add(p)
            if len(result) > result_len:
                result_len += 1
                yield p


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def filled(points):
    """
    Generator function which yields the coordinates necessary to fill the space
    enclosed by the specified *points*.

    This function can be applied to anything that returns a sequence of points.
    For example, to create a filled triangle::

        >>> triangle = [O, 4*X, 4*Z]
        >>> list(filled(lines(triangle)))
        [Vector(x=0, y=0, z=0), Vector(x=0, y=0, z=1), Vector(x=0, y=0, z=2),
         Vector(x=0, y=0, z=3), Vector(x=0, y=0, z=4), Vector(x=1, y=0, z=2),
         Vector(x=1, y=0, z=1), Vector(x=1, y=0, z=0), Vector(x=1, y=0, z=3),
         Vector(x=2, y=0, z=1), Vector(x=2, y=0, z=0), Vector(x=2, y=0, z=2),
         Vector(x=3, y=0, z=1), Vector(x=3, y=0, z=0), Vector(x=4, y=0, z=0)]

    Or to create a filled circle::

        >>> list(filled(circle(O, 4*X)))
        [Vector(x=-4, y=0, z=0), Vector(x=-3, y=-1, z=0), Vector(x=-3, y=-2, z=0),
         Vector(x=-3, y=0, z=0), Vector(x=-3, y=1, z=0), Vector(x=-3, y=2, z=0),
         Vector(x=-2, y=-1, z=0), Vector(x=-2, y=-2, z=0), Vector(x=-2, y=-3, z=0),
         Vector(x=-2, y=0, z=0), Vector(x=-2, y=1, z=0), Vector(x=-2, y=2, z=0),
         Vector(x=-2, y=3, z=0), Vector(x=-1, y=0, z=0), Vector(x=-1, y=-1, z=0),
         Vector(x=-1, y=-2, z=0), Vector(x=-1, y=-3, z=0), Vector(x=-1, y=1, z=0),
         Vector(x=-1, y=2, z=0), Vector(x=-1, y=3, z=0), Vector(x=0, y=-1, z=0),
         Vector(x=0, y=-2, z=0), Vector(x=0, y=-3, z=0), Vector(x=0, y=-4, z=0),
         Vector(x=0, y=0, z=0), Vector(x=0, y=1, z=0), Vector(x=0, y=2, z=0),
         Vector(x=0, y=3, z=0), Vector(x=0, y=4, z=0), Vector(x=1, y=0, z=0),
         Vector(x=1, y=-1, z=0), Vector(x=1, y=-2, z=0), Vector(x=1, y=-3, z=0),
         Vector(x=1, y=1, z=0), Vector(x=1, y=2, z=0), Vector(x=1, y=3, z=0),
         Vector(x=2, y=0, z=0), Vector(x=2, y=-1, z=0), Vector(x=2, y=-2, z=0),
         Vector(x=2, y=-3, z=0), Vector(x=2, y=1, z=0), Vector(x=2, y=2, z=0),
         Vector(x=2, y=3, z=0), Vector(x=3, y=0, z=0), Vector(x=3, y=-1, z=0),
         Vector(x=3, y=-2, z=0), Vector(x=3, y=1, z=0), Vector(x=3, y=2, z=0),
         Vector(x=4, y=0, z=0), Vector(x=4, y=-1, z=0), Vector(x=4, y=1, z=0)]

    To draw the resulting filled object you can simply assign a block to the
    collection of vectors generated (or assign a sequence of blocks of equal
    length if you want the object to have varying block types)::

        >>> world.blocks[filled(lines(triangle))] = Block('stone')

    A simple brute-force algorithm is used that simply generates all the lines
    connecting all specified points. However, duplicate elimination is used to
    ensure that no point within the filled space is yielded twice.

    Note that if you pass the coordinates of a polyhedron which contains holes
    or gaps compared to its convex hull, this function *may* fill those holes
    or gaps (but it will depend on the orientation of the object).
    """
    result = set()
    result_len = 0
    for p, q in pairwise(sorted(points)):
        for l in line(p, q):
            result.add(l)
            if len(result) > result_len:
                result_len += 1
                yield l




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
# Copyright (c) 2012 Dan Crosta
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Compatibility routines to make Python 2 more like Python 3
"""

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')

import sys

# Python 2's xrange is rubbish compared to Python 3's range. The Python 2
# version doesn't permit slicing (which we need for vector_range), and its
# membership test operates in O(n) time (where n is the virtual length of the
# range) instead of O(1) as in Python 3.
#
# Argh! Python 3.2's range is crap as well. Equality tests are completely
# broken and I can't fix them easily because it lacks the start/stop/step
# properties introduced in 3.3. Eurgh...
#
# The following class is partially based on Dan Crosta's wonderful pure Python
# xrange class which can be found at https://github.com/dcrosta/xrange

if sys.version_info < (3, 3):
    from collections import Sequence

    class range(Sequence):
        def __init__(self, start, stop=None, step=1):
            if stop is None:
                start, stop = 0, start
            if (start != start // 1) or (stop != stop // 1) or (step != step // 1):
                raise TypeError('integer arguments are required')
            if step == 0:
                raise ValueError('step must not be zero')
            elif step < 0:
                stop = min(stop, start)
            else:
                stop = max(stop, start)
            self._start = start
            self._stop = stop
            self._step = step
            q, r = divmod(self.stop - self.start, self.step)
            self._len = q + (1 if r else 0)

        @property
        def start(self):
            return self._start

        @property
        def stop(self):
            return self._stop

        @property
        def step(self):
            return self._step

        def __repr__(self):
            if self.step == 1:
                return 'range(%d, %d)' % (self.start, self.stop)
            else:
                return 'range(%d, %d, %d)' % (self.start, self.stop, self.step)

        def __eq__(self, other):
            if isinstance(other, range):
                if len(self) == 0:
                    return len(other) == 0
                elif len(self) == len(other) == 1:
                    return self.start == other.start
                else:
                    # If the other object is a (non-degenerate) range with
                    # equivalent start, last value and step, it's equal. Note
                    # that it's not enough to just test stop here as range(0,
                    # 10, 3) == range(0, 11, 3).
                    self_last = self.start + (len(self) * self.step)
                    other_last = other.start + (len(other) * other.step)
                    return (
                            self.start == other.start and
                            self_last == other_last and
                            self.step == other.step
                            )
            return False

        def __ne__(self, other):
            return not self.__eq__(other)

        def __len__(self):
            return self._len

        def __nonzero__(self):
            return len(self) > 0

        def index(self, value):
            q, r = divmod(value - self.start, self.step)
            if r == 0 and (0 <= q < len(self)):
                return abs(q)
            raise ValueError('%r is not in range' % value)

        def count(self, value):
            return 1 if value in self else 0

        def __contains__(self, value):
            try:
                self.index(value)
            except ValueError:
                return False
            else:
                return True

        def __iter__(self):
            count = 0
            value = self.start
            while count < len(self):
                yield value
                count += 1
                value += self.step

        def __reversed__(self):
            last = self.start + ((len(self) - 1) * self.step)
            return range(last, self.start - self.step, -self.step)

        def __getitem__(self, index):
            if isinstance(index, slice):
                return self._get_slice(index)
            if index < 0:
                # negative indexes access from the end
                index += len(self)
            if not (0 <= index < len(self)):
                raise IndexError('range object index out of range')
            return self.start + index * self.step

        def _get_slice(self, s):
            start, stop, step = s.start, s.stop, s.step

            # Calculate the step of the new range
            if step == 0:
                raise ValueError('slice step cannot be 0')
            elif step is None:
                step = 1
            range_step = step * self.step

            # Calculate the start of the new range
            last = self.start + ((len(self) - 1) * self.step)
            if start is None:
                if step > 0:
                    range_start = self.start
                else:
                    range_start = last
            else:
                if start < 0:
                    start = max(0, start + len(self))
                if start < len(self):
                    range_start = self[start]
                else:
                    range_start = last + self.step

            # Calculate the stop of the new range
            last += self.step
            if stop is None:
                if step > 0:
                    range_stop = last
                else:
                    range_stop = self.start - self.step
            else:
                if stop < 0:
                    stop = max(0, stop + len(self))
                if stop < len(self):
                    range_stop = self[stop]
                else:
                    range_stop = last

            return range(range_start, range_stop, range_step)
else:
    range = range



# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
The player module defines the :class:`Players` class, which is available via
the :attr:`~picraft.world.World.players` attribute, the :class:`Player` class,
which represents an arbitrary player in the world, and the :class:`HostPlayer`
class which represents the player on the host machine (accessible via the
:attr:`~picraft.world.World.player` attribute).

.. note::

    All items in this module are available from the :mod:`picraft` namespace
    without having to import :mod:`picraft.player` directly.

The following items are defined in the module:


Player
======

.. autoclass:: Player
    :inherited-members:
    :members:


HostPlayer
==========

.. autoclass:: HostPlayer
    :inherited-members:
    :members:
"""

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')


from .exc import ConnectionError, NotSupported
from .vector import Vector


class Players(object):
    """
    Thie class implements the :attr:`~picraft.world.World.players` attribute.
    """

    def __init__(self, connection):
        self._connection = connection
        self._cache = {}

    def __repr__(self):
        self._refresh()
        return '<Players keys={%s}>' % (', '.join(str(i) for i in self._cache))

    def _refresh(self):
        self._cache = {
            pid: self._cache.get(pid, Player(self._connection, pid))
            for pid in (
                int(i) for i in
                self._connection.transact('world.getPlayerIds()').split('|')
                )
            }

    def __len__(self):
        self._refresh()
        return len(self._cache)

    def __contains__(self, key):
        self._refresh()
        return key in self._cache

    def __iter__(self):
        self._refresh()
        return iter(self._cache)

    def __getitem__(self, key):
        self._refresh()
        try:
            return self._cache[key]
        except KeyError as e:
            if self._connection.server_version == 'raspberry-juice':
                try:
                    key = int(self._connection.transact('world.getPlayerId(%s)' % key))
                except ConnectionError:
                    # Ignore failed lookups and fall-through to the re-raise
                    pass
                else:
                    return self._cache[key]
            raise e

    def keys(self):
        self._refresh()
        return self._cache.keys()

    def values(self):
        self._refresh()
        return self._cache.values()

    def items(self):
        self._refresh()
        return self._cache.items()


class BasePlayer(object):
    """
    Base class for players.
    """

    def __init__(self, connection, prefix, player_id):
        self._connection = connection
        self._player_id = player_id
        self._prefix = prefix

    def _cmd(self, command, *args):
        if self._player_id is not None:
            args = (self._player_id,) + args
        args = ','.join(str(arg) for arg in args)
        return '%s.%s(%s)' % (self._prefix, command, args)

    def _get_pos(self):
        return Vector.from_string(
            self._connection.transact(self._cmd('getPos')), type=float)
    def _set_pos(self, value):
        self._connection.send(
            self._cmd('setPos', value.x, value.y, value.z))
    pos = property(_get_pos, _set_pos, doc="""\
        The precise position of the player within the world.

        This property returns the position of the selected player within the
        Minecraft world, as a :class:`Vector` instance. This is the *precise*
        position of the player including decimal places (representing portions
        of a tile). You can assign to this property to reposition the player.
        """)

    def _get_tile_pos(self):
        return Vector.from_string(
            self._connection.transact(self._cmd('getTile')))
    def _set_tile_pos(self, value):
        self._connection.send(
            self._cmd('setTile', value.x, value.y, value.z))
    tile_pos = property(_get_tile_pos, _set_tile_pos, doc="""\
        The position of the player within the world to the nearest block.

        This property returns the position of the selected player in the
        Minecraft world to the nearest block, as a :class:`Vector` instance.
        You can assign to this property to reposition the player.
        """)

    @property
    def heading(self):
        """
        The direction the player is facing in clockwise degrees from South.

        This property can be queried to determine the direction that the player
        is facing. The value is returned as a floating-point number of degrees
        from North (i.e. 180 is North, 270 is East, 0 is South, and 90 is
        West).

        .. warning::

            Player heading is only supported on Raspberry Juice.
        """
        if self._connection.server_version != 'raspberry-juice':
            raise NotSupported(
                'cannot query heading on server version: %s' %
                self._connection.server_version)
        return float(
            self._connection.transact(self._cmd('getRotation')))

    @property
    def pitch(self):
        """
        The elevation of the player's view in degrees from the horizontal.

        This property can be queried to determine whether the player is looking
        up (values from 0 to -90) or down (values from 0 down to 90). The value
        is returned as floating-point number of degrees from the horizontal.

        .. warning::

            Player pitch is only supported on Raspberry Juice.
        """
        if self._connection.server_version != 'raspberry-juice':
            raise NotSupported(
                'cannot query pitch on server version: %s' %
                self._connection.server_version)
        return float(
            self._connection.transact(self._cmd('getPitch')))

    @property
    def direction(self):
        """
        The direction the player is facing as a unit vector.

        This property can be queried to retrieve a unit
        :class:`~picraft.vector.Vector` pointing in the direction of the
        player's view.

        .. warning::

            Player direction is only supported on Raspberry Juice.
        """
        if self._connection.server_version != 'raspberry-juice':
            raise NotSupported(
                'cannot query direction on server version: %s' %
                self._connection.server_version)
        return Vector.from_string(
            self._connection.transact(self._cmd('getDirection')),
            type=float)


class Player(BasePlayer):
    """
    Represents a player within the game world.

    Players are uniquely identified by their *player_id*. Instances of this
    class are available from the :attr:`~picraft.world.World.players` mapping.
    It provides properties to query and manipulate the position and settings of
    the player.
    """

    def __init__(self, connection, player_id):
        super(Player, self).__init__(connection, 'entity', player_id)

    def __repr__(self):
        return '<Player player_id=%d>' % self._player_id

    @property
    def player_id(self):
        """
        Returns the integer ID of the player on the server.
        """
        return self._player_id



class HostPlayer(BasePlayer):
    """
    Represents the host player within the game world.

    An instance of this class is accessible as the :attr:`Game.player`
    attribute. It provides properties to query and manipulate the position
    and settings of the host player.
    """

    def __init__(self, connection):
        super(HostPlayer, self).__init__(connection, 'player', None)

    def __repr__(self):
        return '<HostPlayer>'

    def _get_autojump(self):
        raise AttributeError(
                'reading autojump is not supported by the server')
    def _set_autojump(self, value):
        if self._connection.server_version != 'minecraft-pi':
            raise NotSupported(
                'cannot change player settings on server version: %s' %
                self._connection.server_version)
        self._connection.send('player.setting(autojump,%d)' % bool(value))
    autojump = property(_get_autojump, _set_autojump, doc="""\
        Write-only property which sets whether the host player autojumps.

        When this property is set to True (which is the default), the host
        player will automatically jump onto blocks when it runs into them
        (unless the blocks are too high to jump onto).

        .. warning::

            Player settings are only supported on Minecraft Pi edition.

        .. note::

            Unfortunately, the underlying protocol provides no means of reading
            a world setting, so this property is write-only (attempting to
            query it will result in an :exc:`AttributeError` being raised).
        """)




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
The events module defines the :class:`Events` class, which provides methods for
querying events in the Minecraft world, and the :class:`BlockHitEvent`,
:class:`PlayerPosEvent`, :class:`ChatPostEvent`, and :class:`IdleEvent` classes
which represent the various event types.

.. note::

    All items in this module are available from the :mod:`picraft` namespace
    without having to import :mod:`picraft.events` directly.

The following items are defined in the module:


Events
======

.. autoclass:: Events
    :members:


BlockHitEvent
=============

.. autoclass:: BlockHitEvent(pos, face, player)
    :members:


PlayerPosEvent
==============

.. autoclass:: PlayerPosEvent(old_pos, new_pos, player)
    :members:


ChatPostEvent
=============

.. autoclass:: ChatPostEvent(message, player)
    :members:


IdleEvent
=========

.. autoclass:: IdleEvent()
    :members:
"""

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')


import logging
import threading
import time
from collections import namedtuple, Container
from weakref import WeakSet
from functools import update_wrapper
from types import FunctionType

from .exc import ConnectionClosed, NoHandlersWarning
from .vector import Vector
from .player import Player

logger = logging.getLogger('picraft')


class BlockHitEvent(namedtuple('BlockHitEvent', ('pos', 'face', 'player'))):
    """
    Event representing a block being hit by a player.

    This tuple derivative represents the event resulting from a player striking
    a block with their sword in the Minecraft world. Users will not normally
    need to construct instances of this class, rather they are constructed and
    returned by calls to :meth:`~Events.poll`.

    .. note::

        Please note that the block hit event only registers when the player
        *right clicks* with the sword. For some reason, left clicks do not
        count.

    .. attribute:: pos

        A :class:`~picraft.vector.Vector` indicating the position of the block
        which was struck.

    .. attribute:: face

        A string indicating which side of the block was struck. This can be one
        of six values: 'x+', 'x-', 'y+', 'y-', 'z+', or 'z-'. The value
        indicates the axis, and direction along that axis, that the side faces:

        .. image:: block_faces.*

    .. attribute:: player

        A :class:`~picraft.player.Player` instance representing the player that
        hit the block.
    """

    @classmethod
    def from_string(cls, connection, s):
        v, f, p = s.rsplit(',', 2)
        return cls(Vector.from_string(v), {
            0: 'y-',
            1: 'y+',
            2: 'z-',
            3: 'z+',
            4: 'x-',
            5: 'x+',
            }[int(f)], Player(connection, int(p)))

    @property
    def __dict__(self):
        # Ensure __dict__ property works in Python 3.3 and above.
        return super(BlockHitEvent, self).__dict__

    def __repr__(self):
        return '<BlockHitEvent pos=%s face=%r player=%d>' % (
                self.pos, self.face, self.player.player_id)


class PlayerPosEvent(namedtuple('PlayerPosEvent', ('old_pos', 'new_pos', 'player'))):
    """
    Event representing a player moving.

    This tuple derivative represents the event resulting from a player moving
    within the Minecraft world. Users will not normally need to construct
    instances of this class, rather they are constructed and returned by calls
    to :meth:`~Events.poll`.

    .. attribute:: old_pos

        A :class:`~picraft.vector.Vector` indicating the location of the player
        prior to this event. The location includes decimal places (it is not
        the tile-position, but the actual position).

    .. attribute:: new_pos

        A :class:`~picraft.vector.Vector` indicating the location of the player
        as of this event. The location includes decimal places (it is not
        the tile-position, but the actual position).

    .. attribute:: player

        A :class:`~picraft.player.Player` instance representing the player that
        moved.
    """

    @property
    def __dict__(self):
        # Ensure __dict__ property works in Python 3.3 and above.
        return super(PlayerPosEvent, self).__dict__

    def __repr__(self):
        return '<PlayerPosEvent old_pos=%s new_pos=%s player=%d>' % (
                self.old_pos, self.new_pos, self.player.player_id)


class ChatPostEvent(namedtuple('ChatPostEvent', ('message', 'player'))):
    """
    Event representing a chat post.

    This tuple derivative represents the event resulting from a chat message
    being posted in the Minecraft world. Users will not normally need to
    construct instances of this class, rather they are constructed and returned
    by calls to :meth:`~Events.poll`.

    .. note::

        Chat events are only generated by the Raspberry Juice server, not by
        Minecraft Pi edition.

    .. attribute:: message

        The message that was posted to the world.

    .. attribute:: player

        A :class:`~picraft.player.Player` instance representing the player that
        moved.
    """

    @classmethod
    def from_string(cls, connection, s):
        p, m = s.split(',', 1)
        return cls(m, Player(connection, int(p)))

    @property
    def __dict__(self):
        # Ensure __dict__ property works in Python 3.3 and above.
        return super(ChatPostEvent, self).__dict__

    def __repr__(self):
        return '<ChatPostEvent message=%s player=%d>' % (
                self.message, self.player.player_id)


class IdleEvent(namedtuple('IdleEvent', ())):
    """
    Event that fires in the event that no other events have occurred since the
    last poll. This is only used if :attr:`Events.include_idle` is ``True``.
    """

    @property
    def __dict__(self):
        # Ensure __dict__ property works in Python 3.3 and above.
        return super(IdleEvent, self).__dict__

    def __repr__(self):
        return '<IdleEvent>'


class Events(object):
    """
    This class implements the :attr:`~picraft.world.World.events` attribute.

    There are two ways of responding to picraft's events: the first is to
    :meth:`poll` for them manually, and process each event in the resulting
    list::

        >>> for event in world.events.poll():
        ...     print(repr(event))
        ...
        <BlockHitEvent pos=1,1,1 face="y+" player=1>,
        <PlayerPosEvent old_pos=0.2,1.0,0.7 new_pos=0.3,1.0,0.7 player=1>

    The second is to "tag" functions as event handlers with the decorators
    provided and then call the :meth:`main_loop` function which will handle
    polling the server for you, and call all the relevant functions as needed::

        @world.events.on_block_hit(pos=Vector(1,1,1))
        def hit_block(event):
            print('You hit the block at %s' % event.pos)

        world.events.main_loop()

    By default, only block hit events will be tracked. This is because it is
    the only type of event that the Minecraft server provides information about
    itself, and thus the only type of event that can be processed relatively
    efficiently. If you wish to track player positions, assign a set of player
    ids to the :attr:`track_players` attribute. If you wish to include idle
    events (which fire when nothing else is produced in response to
    :meth:`poll`) then set :attr:`include_idle` to ``True``.

    .. note::

        If you are using a Raspberry Juice server, chat post events are also
        tracked by default. Chat post events are only supported with Raspberry
        Juice servers; Minecraft Pi edition doesn't support chat post events.

    Finally, the :attr:`poll_gap` attribute specifies how long to pause during
    each iteration of :meth:`main_loop` to permit event handlers some time to
    interact with the server. Setting this to 0 will provide the fastest
    response to events, but will result in event handlers having to fight with
    event polling for access to the server.
    """

    def __init__(self, connection, poll_gap=0.1, include_idle=False):
        self._connection = connection
        self._handlers = []
        self._handler_instances = WeakSet()
        self._poll_gap = poll_gap
        self._include_idle = include_idle
        self._track_players = {}

    def _get_poll_gap(self):
        return self._poll_gap
    def _set_poll_gap(self, value):
        self._poll_gap = float(value)
    poll_gap = property(_get_poll_gap, _set_poll_gap, doc="""\
        The length of time (in seconds) to pause during :meth:`main_loop`.

        This property specifies the length of time to wait at the end of each
        iteration of :meth:`main_loop`. By default this is 0.1 seconds.

        The purpose of the pause is to give event handlers executing in the
        background time to communicate with the Minecraft server. Setting this
        to 0.0 will result in faster response to events, but also starves
        threaded event handlers of time to communicate with the server,
        resulting in "choppy" performance.
        """)

    def _get_track_players(self):
        return self._track_players.keys()
    def _set_track_players(self, value):
        try:
            self._track_players = {
                pid: Player(self._connection, pid).pos.round(1)
                for pid in value
                }
        except TypeError:
            if not isinstance(value, int):
                raise ValueError(
                        'track_players value must be a player id '
                        'or a sequence of player ids')
            self._track_players = {
                value: Player(self._connection, value).pos
                }
    track_players = property(_get_track_players, _set_track_players, doc="""\
        The set of player ids for which movement should be tracked.

        By default the :meth:`poll` method will not produce player position
        events (:class:`PlayerPosEvent`). Producing these events requires extra
        interactions with the Minecraft server (one for each player tracked)
        which slow down response to block hit events.

        If you wish to track player positions, set this attribute to the set of
        player ids you wish to track and their positions will be stored.  The
        next time :meth:`poll` is called it will query the positions for all
        specified players and fire player position events if they have changed.

        Given that the :attr:`~picraft.world.World.players` attribute
        represents a dictionary mapping player ids to players, if you wish to
        track all players you can simply do::

            >>> world.events.track_players = world.players
        """)

    def _get_include_idle(self):
        return self._include_idle
    def _set_include_idle(self, value):
        self._include_idle = bool(value)
    include_idle = property(_get_include_idle, _set_include_idle, doc="""\
        If ``True``, generate an idle event when no other events would be
        generated by :meth:`poll`. This attribute defaults to ``False``.
        """)

    def clear(self):
        """
        Forget all pending events that have not yet been retrieved with
        :meth:`poll`.

        This method is used to clear the list of events that have occurred
        since the last call to :meth:`poll` without retrieving them. This is
        useful for ensuring that events subsequently retrieved definitely
        occurred *after* the call to :meth:`clear`.
        """
        self._set_track_players(self._get_track_players())
        self._connection.send('events.clear()')

    def poll(self):
        """
        Return a list of all events that have occurred since the last call to
        :meth:`poll`.

        For example::

            >>> w = World()
            >>> w.events.track_players = w.players
            >>> w.events.include_idle = True
            >>> w.events.poll()
            [<PlayerPosEvent old_pos=0.2,1.0,0.7 new_pos=0.3,1.0,0.7 player=1>,
             <BlockHitEvent pos=1,1,1 face="x+" player=1>,
             <BlockHitEvent pos=1,1,1 face="x+" player=1>]
            >>> w.events.poll()
            [<IdleEvent>]
        """
        def player_pos_events(positions):
            for pid, old_pos in positions.items():
                player = Player(self._connection, pid)
                new_pos = player.pos.round(1)
                if old_pos != new_pos:
                    yield PlayerPosEvent(old_pos, new_pos, player)
                positions[pid] = new_pos

        def block_hit_events():
            s = self._connection.transact('events.block.hits()')
            if s:
                for e in s.split('|'):
                    yield BlockHitEvent.from_string(self._connection, e)

        def chat_post_events():
            if self._connection.server_version == 'raspberry-juice':
                s = self._connection.transact('events.chat.posts()')
                if s:
                    for e in s.split('|'):
                        yield ChatPostEvent.from_string(self._connection, e)

        events = list(player_pos_events(self._track_players)) + list(block_hit_events()) + list(chat_post_events())

        if events:
            return events
        elif self._include_idle:
            return [IdleEvent()]
        else:
            return []

    def main_loop(self):
        """
        Starts the event polling loop when using the decorator style of event
        handling (see :meth:`on_block_hit`).

        This method will not return, so be sure that you have specified all
        your event handlers before calling it. The event loop can only be
        broken by an unhandled exception, or by closing the world's connection
        (in the latter case the resulting :exc:`~picraft.exc.ConnectionClosed`
        exception will be suppressed as it is assumed that you want to end the
        script cleanly).
        """
        logger.info('Entering event loop')
        try:
            while True:
                self.process()
                time.sleep(self.poll_gap)
        except ConnectionClosed:
            logger.info('Connection closed; exiting event loop')

    def process(self):
        """
        Poll the server for events and call any relevant event handlers
        registered with :meth:`on_block_hit`.

        This method is called repeatedly the event handler loop implemented by
        :meth:`main_loop`; developers should only call this method when
        implementing their own event loop manually, or when their (presumably
        non-threaded) event handler is engaged in a long operation and they
        wish to permit events to be processed in the meantime.
        """
        for event in self.poll():
            for handler in self._handlers:
                if handler.matches(event):
                    handler.execute(event)

    def has_handlers(self, cls):
        """
        Decorator for registering a class as containing picraft event handlers.

        If you are writing a class which contains methods that you wish to
        use as event handlers for picraft events, you must decorate the class
        with ``@has_handlers``. This will ensure that picraft tracks instances
        of the class and dispatches events to each instance that exists when
        the event occurs.

        For example::

            from picraft import World, Block, Vector, X, Y, Z

            world = World()

            @world.events.has_handlers
            class HitMe(object):
                def __init__(self, pos):
                    self.pos = pos
                    self.been_hit = False
                    world.blocks[self.pos] = Block('diamond_block')

                @world.events.on_block_hit()
                def was_i_hit(self, event):
                    if event.pos == self.pos:
                        self.been_hit = True
                        print('Block at %s was hit' % str(self.pos))

            p = world.player.tile_pos
            block1 = HitMe(p + 2*X)
            block2 = HitMe(p + 2*Z)
            world.events.main_loop()

        Class-based handlers are an advanced feature and have some notable
        limitations. For instance, in the example above the ``on_block_hit``
        handler couldn't be declared with the block's position because this was
        only known at instance creation time, not at class creation time (which
        was when the handler was registered).

        Furthermore, class-based handlers must be regular instance methods
        (those which accept the instance, self, as the first argument); they
        cannot be class methods or static methods.

        .. note::

            The ``@has_handlers`` decorator takes no arguments and shouldn't
            be called, unlike event handler decorators.
        """
        # Search the class for handler methods, appending the class to the
        # handler's list of associated classes (if you're thinking why is this
        # a collection, consider that a method can be associated with multiple
        # classes either by inheritance or direct assignment)
        handlers_found = 0
        for item in dir(cls):
            item = getattr(cls, item, None)
            if item: # PY2
                item = getattr(item, 'im_func', item)
            if item and isinstance(item, FunctionType):
                try:
                    item._picraft_classes.add(cls)
                    handlers_found += 1
                except AttributeError:
                    pass
        if not handlers_found:
            warnings.warn(NoHandlersWarning('no handlers found in %s' % cls))
            return cls
        # Replace __init__ on the class with a closure that adds every instance
        # constructed to self._handler_instances. As this is a WeakSet,
        # instances that die will be implicitly removed
        old_init = getattr(cls, '__init__', None)
        def __init__(this, *args, **kwargs):
            if old_init:
                old_init(this, *args, **kwargs)
            self._handler_instances.add(this)
        if old_init:
            update_wrapper(__init__, old_init)
        cls.__init__ = __init__
        return cls

    def _handler_closure(self, f):
        def handler(event):
            if not f._picraft_classes:
                # The handler is a straight-forward function; just call it
                f(event)
            else:
                # The handler is an unbound method (yes, I know these don't
                # really exist in Python 3; it's a function which is expecting
                # to be called from an object instance if you like). Here we
                # search the set of instances of classes which were registered
                # as having handlers (by @has_handlers)
                for cls in f._picraft_classes:
                    for inst in self._handler_instances:
                        # Check whether the instance has the right class; note
                        # that we *don't* use isinstance() here as we want an
                        # exact match
                        if inst.__class__ == cls:
                            # Bind the function to the instance via its
                            # descriptor
                            f.__get__(inst, cls)(event)
        update_wrapper(handler, f)
        return handler

    def on_idle(self, thread=False, multi=True):
        """
        Decorator for registering a function/method as an idle handler.

        This decorator is used to mark a function as an event handler which
        will be called when no other event handlers have been called in an
        iteration of :meth:`main_loop`. The function will be called with the
        corresponding :class:`IdleEvent` as the only argument.

        Note that idle events will only be generated if :attr:`include_idle`
        is set to ``True``.
        """
        def decorator(f):
            self._handlers.append(
                    IdleHandler(self._handler_closure(f), thread, multi))
            f._picraft_classes = set()
            return f
        return decorator

    def on_player_pos(self, thread=False, multi=True, old_pos=None, new_pos=None):
        """
        Decorator for registering a function/method as a position change
        handler.

        This decorator is used to mark a function as an event handler which
        will be called for any events indicating that a player's position has
        changed while :meth:`main_loop` is executing. The function will be
        called with the corresponding :class:`PlayerPosEvent` as the only
        argument.

        The *old_pos* and *new_pos* parameters can be used to specify vectors
        or sequences of vectors (including a
        :class:`~picraft.vector.vector_range`) that the player position events
        must match in order to activate the associated handler. For example, to
        fire a handler every time any player enters or walks over blocks within
        (-10, 0, -10) to (10, 0, 10)::

            from picraft import World, Vector, vector_range

            world = World()
            world.events.track_players = world.players

            from_pos = Vector(-10, 0, -10)
            to_pos = Vector(10, 0, 10)
            @world.events.on_player_pos(new_pos=vector_range(from_pos, to_pos + 1))
            def in_box(event):
                world.say('Player %d stepped in the box' % event.player.player_id)

            world.events.main_loop()

        Various effects can be achieved by combining *old_pos* and *new_pos*
        filters. For example, one could detect when a player crosses a boundary
        in a particular direction, or decide when a player enters or leaves a
        particular area.

        Note that only players specified in :attr:`track_players` will generate
        player position events.
        """
        def decorator(f):
            self._handlers.append(
                    PlayerPosHandler(self._handler_closure(f),
                        thread, multi, old_pos, new_pos))
            f._picraft_classes = set()
            return f
        return decorator

    def on_block_hit(self, thread=False, multi=True, pos=None, face=None):
        """
        Decorator for registering a function/method as a block hit handler.

        This decorator is used to mark a function as an event handler which
        will be called for any events indicating a block has been hit while
        :meth:`main_loop` is executing. The function will be called with the
        corresponding :class:`BlockHitEvent` as the only argument.

        The *pos* parameter can be used to specify a vector or sequence of
        vectors (including a :class:`~picraft.vector.vector_range`); in this
        case the event handler will only be called for block hits on matching
        vectors.

        The *face* parameter can be used to specify a face or sequence of
        faces for which the handler will be called.

        For example, to specify that one handler should be called for hits
        on the top of any blocks, and another should be called only for hits
        on any face of block at the origin one could use the following code::

            from picraft import World, Vector

            world = World()

            @world.events.on_block_hit(pos=Vector(0, 0, 0))
            def origin_hit(event):
                world.say('You hit the block at the origin')

            @world.events.on_block_hit(face="y+")
            def top_hit(event):
                world.say('You hit the top of a block at %d,%d,%d' % event.pos)

            world.events.main_loop()

        The *thread* parameter (which defaults to ``False``) can be used to
        specify that the handler should be executed in its own background
        thread, in parallel with other handlers.

        Finally, the *multi* parameter (which only applies when *thread* is
        ``True``) specifies whether multi-threaded handlers should be allowed
        to execute in parallel. When ``True`` (the default), threaded handlers
        execute as many times as activated in parallel. When ``False``, a
        single instance of a threaded handler is allowed to execute at any
        given time; simultaneous activations are ignored (but not queued, as
        with unthreaded handlers).
        """
        def decorator(f):
            self._handlers.append(
                    BlockHitHandler(self._handler_closure(f),
                        thread, multi, pos, face))
            f._picraft_classes = set()
            return f
        return decorator

    def on_chat_post(self, thread=False, multi=True, message=None):
        """
        Decorator for registering a function/method as a chat event handler.

        This decorator is used to mark a function as an event handler which
        will be called for events indicating a chat message was posted to
        the world while :meth:`main_loop` is executing. The function will be
        called with the corresponding :class:`ChatPostEvent` as the only
        argument.

        .. note::

            Only the Raspberry Juice server generates chat events; Minecraft
            Pi Edition does not support this event type.

        The *message* parameter can be used to specify a string or regular
        expression; in this case the event handler will only be called for chat
        messages which match this value. For example::

            import re
            from picraft import World, Vector

            world = World()

            @world.events.on_chat_post(message="hello world")
            def echo(event):
                world.say("Hello player %d!" % event.player.player_id)

            @world.events.on_chat_post(message=re.compile(r"teleport_me \d+,\d+,\d+"))
            def teleport(event):
                x, y, z = event.message[len("teleport_me "):].split(",")
                event.player.pos = Vector(int(x), int(y), int(z))

            world.events.main_loop()

        The *thread* parameter (which defaults to ``False``) can be used to
        specify that the handler should be executed in its own background
        thread, in parallel with other handlers.

        Finally, the *multi* parameter (which only applies when *thread* is
        ``True``) specifies whether multi-threaded handlers should be allowed
        to execute in parallel. When ``True`` (the default), threaded handlers
        execute as many times as activated in parallel. When ``False``, a
        single instance of a threaded handler is allowed to execute at any
        given time; simultaneous activations are ignored (but not queued, as
        with unthreaded handlers).
        """
        def decorator(f):
            self._handlers.append(
                    ChatPostHandler(self._handler_closure(f),
                        thread, multi, message))
            f._picraft_classes = set()
            return f
        return decorator


class EventHandler(object):
    """
    This is an internal object used to associate event handlers with their
    activation restrictions.

    The *action* parameter specifies the function to be run when a matching
    event is received from the server.

    The *thread* parameter specifies whether the *action* will be launched in
    its own background thread. If *multi* is ``False``, then the
    :meth:`execute` method will ensure that any prior execution has finished
    before launching another one.
    """

    def __init__(self, action, thread, multi):
        self.action = action
        self.thread = thread
        self.multi = multi
        self._thread = None

    def execute(self, event):
        """
        Launches the *action* in a background thread if necessary. If required,
        this method also ensures threaded actions don't overlap.
        """
        if self.thread:
            if self.multi:
                threading.Thread(target=self._execute_handler, args=(event,)).start()
            elif not self._thread:
                self._thread = threading.Thread(target=self._execute_single, args=(event,))
                self._thread.start()
        else:
            self._execute_handler(event)

    def _execute_single(self, event):
        try:
            self._execute_handler(event)
        finally:
            self._thread = None

    def _execute_handler(self, event):
        self.action(event)

    def matches(self, event):
        """
        Tests whether or not *event* match all the filters for the handler that
        this object represents.
        """
        raise NotImplementedError


class PlayerPosHandler(EventHandler):
    """
    This class associates a handler with a player-position event.

    Constructor parameters are similar to the parent class,
    :class:`EventHandler` but additionally include *old_pos* and *new_pos* to
    specify the vectors (or sequences of vectors) that an event must transition
    across in order to activate this action. These filters must both match in
    order for the action to fire.
    """

    def __init__(self, action, thread, multi, old_pos, new_pos):
        super(PlayerPosHandler, self).__init__(action, thread, multi)
        self.old_pos = old_pos
        self.new_pos = new_pos

    def matches(self, event):
        return (
                isinstance(event, PlayerPosEvent) and
                self.matches_pos(self.old_pos, event.old_pos) and
                self.matches_pos(self.new_pos, event.new_pos))

    def matches_pos(self, test, pos):
        if test is None:
            return True
        if isinstance(test, Vector):
            return test == pos.floor()
        if isinstance(test, Container):
            return pos.floor() in test
        raise TypeError(
                "%r is not a valid position test; expected Vector or "
                "sequence of Vector" % test)


class BlockHitHandler(EventHandler):
    """
    This class associates a handler with a block-hit event.

    Constructor parameters are similar to the parent class,
    :class:`EventHandler` but additionally include *pos* to specify the vector
    (or sequence of vectors) which an event must match in order to activate
    this action, and *face* to specify the block face (or set of faces) which
    an event must match.  These filters must both match in order for the action
    to fire.
    """

    def __init__(self, action, thread, multi, pos, face):
        super(BlockHitHandler, self).__init__(action, thread, multi)
        self.pos = pos
        if isinstance(face, bytes):
            face = face.decode('ascii')
        self.face = face

    def matches(self, event):
        return (
                isinstance(event, BlockHitEvent) and
                self.matches_pos(event.pos) and
                self.matches_face(event.face))

    def matches_pos(self, pos):
        if self.pos is None:
            return True
        if isinstance(self.pos, Vector):
            return self.pos == pos
        if isinstance(self.pos, Container):
            return pos in self.pos
        raise TypeError(
                "%r is not a valid position test; expected Vector or "
                "sequence of Vector" % pos)

    def matches_face(self, face):
        if self.face is None:
            return True
        if isinstance(self.face, str):
            return self.face == face
        if isinstance(self.face, Container):
            return face in self.face
        raise TypeError(
                "%r is not a valid face test; expected string or sequence "
                "of strings" % face)


class ChatPostHandler(EventHandler):
    """
    This class associates a handler with a chat-post event.

    Constructor parameters are similar to the parent class,
    :class:`EventHandler` but additionally include *message* to specify the
    message that an event must contain in order to activate this action.
    """

    def __init__(self, action, thread, multi, message):
        super(ChatPostHandler, self).__init__(action, thread, multi)
        if isinstance(message, bytes):
            message = message.decode('ascii')
        self.message = message

    def matches(self, event):
        return (
            isinstance(event, ChatPostEvent) and
            self.matches_message(event.message))

    def matches_message(self, message):
        if self.message is None:
            return True
        if isinstance(self.message, str):
            return self.message == message
        try:
            return self.message.match(message)
        except AttributeError:
            raise TypeError(
                    "%r is not a valid message test; expected string"
                    "or regular expression" % message)


class IdleHandler(EventHandler):
    """
    This class associates a handler with an idle event.
    """

    def matches(self, event):
        return isinstance(event, IdleEvent)
